#ifndef STRUCTS_H
#define STRUCTS_H
#include <shared_mutex>
#include <string>
#include <boost/log/trivial.hpp>
#include <boost/property_tree/ptree_fwd.hpp>
#include <terminal_api/ControlProtoCInterface.h>


namespace api_driver::obj {
    /**
     * Обертка для объектов, доступных для обновления
     * NOTE: перед вызовом функций, требующих `TSID`, необходимо захватить мютекс API.
     */
    class CpUpdatebleObject {
    public:
        int64_t lastUpdate = 0;
        int64_t updatePeriodMs = -1;

        virtual void updateCallback(TSID sid, CP_Result& lastCpError) = 0;

        bool checkNeedUpdate(int64_t now) const;

        int64_t getNextUpdate(int64_t now) const;

        virtual ~CpUpdatebleObject();
    };


#ifdef MODEM_IS_SCPC
    class api_driver::StatisticsLogger {
    public:
        StatisticsLogger(): timeStart(TIME_NOW()) {}

        int64_t timeStart;

        bool logEn = false;
        std::atomic<int> logPeriodMs = 1000;
        std::atomic<int> maxAgeMs = 10000;

        /**
         *
         * @return {"en": bool, "logPeriodMs": int, "maxAgeMs": int}
         */
        std::string getSettings() {
            std::lock_guard _lock(mutex);
            std::stringstream res;
            res << "{\"en\":" << boolAsStr(this->logEn);
            res << ",\"logPeriodMs\":" << logPeriodMs;
            res << ",\"maxAgeMs\":" << maxAgeMs;
            res << '}';
            return res.str();
        }
        void setSettings(boost::property_tree::ptree &pt) {
            const bool newEn = pt.get<bool>("en");
            const int newInterval = pt.get<int>("logPeriodMs");
            const int newMaxAgeMs = pt.get<int>("maxAgeMs");

            std::lock_guard _lock(this->mutex);
            this->logPeriodMs = newInterval;
            this->maxAgeMs = newMaxAgeMs;

            if (newEn != this->logEn) {
                if (newEn) {
                    this->logFile.open("/tmp/weblog-statistics.csv", std::ios::out);
                    if (this->logFile.is_open()) {
                        const auto* header = "timestamp\tcnt ok\tcnt bad\tfine freq dem\tcrs freq dem\tcrs freq compensator\tcrs time est\tfine time est\tmax level corr\tcurrent delay\tSNR\tcurrent modcod\tfine freq compensator\tind freq grb\tind freq tochn\tind filt adapt\tfilter corr cinc\tcorr cnt\tRSS\tcor erl\tcor lat\tgc gain\tpower pl rx\n";
                        this->logFile.write(header, static_cast<std::streamsize>(strlen(header)));
                        this->logEn = true;
                        this->timeStart = TIME_NOW();
                    }
                } else {
                    if (this->logFile.is_open()) {
                        this->logFile.close();
                    }
                    this->logEn = false;
                }
            }
        }

        /**
         * Записать значение в "базу данных". Метку при этом вставлять не нужно, она будет вставлена автоматически.
         * @param item
         */
        void putItem(const debug_metrics& item) {
            std::lock_guard _lock(this->mutex);
            if (!logEn) return;
            if (this->logFile.is_open()) {
                std::stringstream res;
                res << makeTimepointFromMillis(TIME_NOW()) << '\t';
                res << item.cnt_ok << '\t';
                res << item.cnt_bad << '\t';
                res << item.fine_freq_dem << '\t';
                res << item.crs_freq_dem << '\t';
                res << item.crs_freq_compensator << '\t';
                res << item.crs_time_est << '\t';
                res << item.fine_time_est << '\t';
                res << item.max_level_corr << '\t';
                res << item.current_delay << '\t';
                res << item.SNR << '\t';
                res << item.current_modcod << '\t';
                res << item.fine_freq_compensator << '\t';
                res << item.ind_freq_grb << '\t';
                res << item.ind_freq_tochn << '\t';
                res << item.ind_filt_adapt << '\t';
                res << item.filter_corr_cinc << '\t';
                res << item.corr_cnt << '\t';
                res << item.RSS << '\t';
                res << item.cor_erl << '\t';
                res << item.cor_lat << '\t';
                res << item.gc_gain << '\t';
                res << item.power_pl_rx << '\n';

                const auto out = res.str();
                this->logFile.write(out.c_str(), static_cast<std::streamsize>(out.length()));
                this->logFile.flush();
            }
        }

        // void collectExpiredItems();

        // void resetLogs() {
        //     std::lock_guard _lock(mutex);
        //     logs.clear();
        // }

        ~StatisticsLogger() = default;
    private:
        // std::pmr::deque<LogItem> logs;
        std::fstream logFile{};
        std::shared_mutex mutex;
    };
#endif

#ifdef MODEM_IS_SCPC
    static constexpr const char* DEFAULT_SERVER_NAME = "RCSM-101";
#endif
#ifdef MODEM_IS_TDMA
    static constexpr const char* DEFAULT_SERVER_NAME = "RCSM-101 TDMA";
#endif

#if defined(MODEM_IS_SCPC)
    class TerminalNetworkSettings: public CpUpdatebleObject {
    public:
        std::string managementIp, managementGateway, dataIp, serverName;
        bool isL2 = true;
        unsigned int dataMtu = 1500;

        TerminalNetworkSettings();
        TerminalNetworkSettings(const TerminalNetworkSettings& src);
        TerminalNetworkSettings& operator= (const TerminalNetworkSettings& src);

        void loadDefaults() {
            managementIp = "0.0.0.0";
            managementGateway = "";
            isL2 = true;
            dataIp = "0.0.0.0";
            dataMtu = 1500;
            serverName = DEFAULT_SERVER_NAME;
        }

        void updateCallback(TSID sid, CP_Result &lastCpError) override;
        void updateFromPt(boost::property_tree::ptree &pt);
        void store(TSID sid, CP_Result& lastCpError);
        std::string asJson();

        ~TerminalNetworkSettings() override;
    };
#endif

    class TerminalFirmwareVersion {
    public:
        std::string version, modemId, modemSn, macMang, macData;

        TerminalFirmwareVersion();
        TerminalFirmwareVersion(const TerminalFirmwareVersion& src);
        ~TerminalFirmwareVersion();

        std::string asJson();

        TerminalFirmwareVersion& operator= (const TerminalFirmwareVersion& src);
    };

    /**
     * Обертка состояния терминала, тут состояние девайса, модулятора/демодулятора, состояние автоматического обновления.
     */
    class TerminalState: public CpUpdatebleObject {
    public:
        std::string fInitState{};
        bool fIsTest = false; // daemon->isTest()
        bool fTxState = false;
#ifdef MODEM_IS_SCPC
        bool fIsCinC = false;
#endif

        bool fStatRxState;
        bool fStatRxSymSyncLock;
        bool fStatRxFreqSearchLock;
        bool fStatRxAfcLock;
        bool fStatRxPktSync;

        float fStatRxSnr;
        float fStatRxRssi;
        uint16_t fStatRxModcod;
        bool fStatRxFrameSizeNormal;
        bool fStatRxIsPilots;

        double fStatRxSymError;
        double fStatRxFreqErr;
        double fStatRxFreqErrAcc;
        double fStatRxInputSignalLevel;
        double fStatRxPllError;
        double fStatRxSpeedOnRxKbit;
        double fStatRxSpeedOnIifKbit;
        uint32_t fStatRxPacketsOk;
        uint32_t fStatRxPacketsBad;
        uint32_t fStatRxPacketsDummy;

        bool fStatTxState;
        uint16_t fStatTxModcod;
        double fStatTxSpeedOnTxKbit;
        double fStatTxSpeedOnIifKbit;
#ifdef MODEM_IS_SCPC
        float fStatTxSnr;
        bool fStatTxFrameSizeNormal;
        bool fStatTxIsPilots;
        double fStatCincOcc;
        bool fStatCincCorrelator;
        uint32_t fStatCincCorrelatorFails;
        int32_t fStatCincFreqErr;
        int32_t fStatCincFreqErrAcc;
        float fStatCincChannelDelay;
#endif
#ifdef MODEM_IS_TDMA
        fStatTxCenterFreq;
        fStatTxSymSpeed;
#endif

        /**
         * Обновление основной части статистики, то есть RX/TX и sysinfo
         */
        void updateCallback(TSID sid, CP_Result &lastCpError) override;

        void updateFromPt(boost::property_tree::ptree &pt);
        void store(TSID sid, CP_Result& lastCpError);
        std::string asJson();

        ~TerminalState() override;
    };

    class TerminalDeviceState: public CpUpdatebleObject {
    public:
        time_t fStatOsUptime;
        double fStatOsLoad1;
        double fStatOsLoad5;
        double fStatOsLoad15;
        int fStatOsTotalram;
        int fStatOsFreeram;
        int fStatOsProcs;

        double fStatDeviceAdrv;
        double fStatDeviceZynq;
        double fStatDeviceFpga;
#ifdef MODEM_IS_TDMA
        DOWNLOAD_STATUS fStatDeviceUpgradeStatus;
        unsigned int fStatDeviceUpgradePercent;
        std::string fStatDeviceUpgradeImage;
#endif

        void updateCallback(TSID sid, CP_Result &lastCpError) override;
        void updateFromPt(boost::property_tree::ptree &pt);
        void store(TSID sid, CP_Result& lastCpError);
        std::string asJson();

        ~TerminalDeviceState() override;
    };
}

#endif //STRUCTS_H
